# faces2DXFentities.rb - Renaud ILTIS - september 2016 - www.cncfab.fr
# 
# SketchUp face to DXF entities converter, with Arc, Circles and polylines entities
#
# Information and credits : see help_text in the code or the Help menu of the plugin in SketchUp
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

require('sketchup.rb')

module CNCFAB

	module ImportExport

	# ******* Export to DXF with entities Arc and Circle *********
	def self.export_dialog(current_skp_lang)
		 if current_skp_lang == "fr"
				result = Sketchup.set_status_text "Nombre de faces traitées : ", SB_VCB_LABEL
				result = Sketchup.set_status_text "",SB_VCB_VALUE
				result = Sketchup.set_status_text "Explications disponibles dans le menu : Extensions > Export Faces to DXF entitites > Help", SB_PROMPT
		 else
				result = Sketchup.set_status_text "Number of treated faces : ", SB_VCB_LABEL
				result = Sketchup.set_status_text "",SB_VCB_VALUE
				result = Sketchup.set_status_text "Explanations avalaibles in the menu : Extensions > Export Faces to DXF entitites > Help", SB_PROMPT
		 end
		 cu=Sketchup.active_model.options['UnitsOptions']["LengthUnit"]
		 case cu
		 when 4
				current_skp_lang == "fr" ? current_unit = "Mètres" : current_unit = "Meters"
		 when 3
				current_skp_lang == "fr" ? current_unit = "Centimètres" : current_unit= "Centimeters"
		 when 2
				current_skp_lang == "fr" ? current_unit = "Millimètres" : current_unit= "Millimeters"
		 when 1
				current_skp_lang == "fr" ? current_unit = "Pieds" : current_unit= "Feet"
		 when 0
				current_skp_lang == "fr" ? current_unit= "Pouces" : current_unit= "Inches"
		 end
		 if current_skp_lang == "fr"
				units_list=["Mètres","Centimètres","Millimètres","Pouces","Pieds"].join("|")
				prompts=["Choisissez l'unité d'export : "]
				enums=[units_list]
				values=[current_unit]
				results = inputbox prompts, values, enums, "Faces to DXF entities"
		 else
				units_list=["Meters","Centimeters","Millimeters","Inches","Feet"].join("|")
				prompts=["Choose the export unit : "]
				enums=[units_list]
				values=[current_unit]
				results = inputbox prompts, values, enums, "Faces to DXF entities"
		 end
		 return if not results
		 case results[0]
		 when "Meters", "Mètres"
				$faces2dxf_conv=0.0254
		 when "Centimeters", "Centimètres"
				$faces2dxf_conv=2.54
		 when "Millimeters", "Millimètres"
				$faces2dxf_conv=25.4
		 when "Feet", "Pieds"
				$faces2dxf_conv=0.0833333333333333
		 when "Inches", "Pouces"
				$faces2dxf_conv=1.0
		 end
	end

	def self.export_faces(z)
		 current_skp_lang = Sketchup.os_language
		 $faces2dxf_conv=25.4  #default unit : mm
		 return if not export_dialog(current_skp_lang)  # give the possibility to change the current unit for the export
		 model = Sketchup.active_model
		 model_filename = File.basename(model.path)
		 if (model_filename == '')
				model_filename = 'model'
		 end
		 selection = model.selection
		 if (selection.empty?)
				selection = model.entities
		 end   
		 if current_skp_lang == "fr"
				out_name = UI.savepanel('Emplacement du fichier dxf à exporter', '' , "#{File.basename(model.path, ".*")}.dxf")
		 else
				out_name = UI.savepanel('Location of the dxf export file', '' , "#{File.basename(model.path, ".*")}.dxf")
		 end
		 begin
			 if out_name && ! out_name.empty?
					if File.extname(out_name).downcase != ".dxf"    # if the user erase the .dxf extension, rewrite it!
						 out_name += ".dxf"
					end
					$number_of_exported_faces = 0
					file = File.new(out_name , 'w')
					write_header(file, selection)
					write_content(file, selection, Geom::Transformation.new(), z)  #Geom::Transformation.new() will be used for the transfert of the coordinates of the groups and components in the Sketchup World Coordinate System (tform)
					write_footer(file, selection)
					file.close()
					current_skp_lang == "fr" ? UI.messagebox("Le fichier dxf est ici : " + out_name) : UI.messagebox("The dxf file is here : " + out_name)
			 else
					# p 'Canceled'
			 end
		 rescue Exception => error
				if current_skp_lang == "fr"
					 UI.messagebox(error.to_s+"\n\nL'export a échoué. Voici quelques causes possibles :\n- Vous voulez écraser un fichier qui est déjà ouvert dans une autre application. => Fermez le fichier ou changez de nom d'export.\n- Vous n'avez pas les droits d'écriture à l'emplacement de l'export. => Augmentez vos droits ou changez d'emplacement.\n- Bug du plugin. => Contactez l'auteur.")
				else
					 UI.messagebox(error.to_s+"\n\nExport failed. Here are some possible causes :\n- You want to overwrite a file that is already open in another application. => Close the file or change the export name.\n- You don't have write access to the location of the export. => Increase your rights or change location.\n- Plugin's bug => Contact the author.")
				end
		 end
	end

	def self.readmefrench
		help_text = "Plugin d'import et d'export par CNCFAB (www.cncfab.fr), fabricant de machines CNC de découpe des polystyrènes par fil chaud.\n==========================================================\n\n* Import faces or polylines from TXT (FilChaudNX => SketchUp):\n-----------------------------------------------------------------------------\nCet outil sert à importer les profils et projets exportés pour SketchUp avec le logiciel de découpe par fil chaud FilChaudNX sous la forme d'une fichier texte (.txt).\n\nIl ouvre un fichier .txt dans lequel chaque ligne contient les coordonnées (x, y, z) d'un point, sous la forme X:Y:Z (attention, séparateur décimal : le point).\n\nPour définir une face, on indique le mot-clé FACE sur une ligne, puis tous les points qui la constituent (un par ligne), et on termine par le mot-clé END_FACE.\n\nPour inverser l'orientation d'une face, on peut terminer la séquence par le mot-clé END_FACE_REVERSE (dans SketchUp, les faces créées dans le plan Z=0 sont toujours orientée vers le bas au moment de la création ; pour les autres faces, cela dépend du sens de rotation des points autour de la face : sens horaire = face arrière).\n\nPour une polyligne, on indique le mot-clé POLYLINE, puis tous les points qui la constituent (un par ligne), et on termine par le mot-clé END_POLYLINE.\n\nLes nombres n'ont pas d'unité, ils seront à l'unité du projet SketchUp au moment de l'import (si on importe 56.3 dans un projet SketchUp en mm, cela fera 56.3mm, si c'est dans un projet en inches, cela fera 56.3inches).\n\n* Import polyline from DAT:\n-----------------------------------------------------------------------------\nCet outil sert à importer un fichier de profil d'aile au format .DAT standard qui sera représenté par une succession de segments avec une corde totale de 100 unités (multiplication par 100 des coordonnées contenues dans le fichier .DAT).\n\n* Export selected faces to X:Y file (SketchUp => FilChaudNX/FilChaud2D):\n-----------------------------------------------------------------------------------------\nCet outil sert à exporter les coordonnées 2D des points des contours des faces sélectionnées pour les importer ensuite dans les logiciels FilChaudNX et FilChaud2D.\n\nChaque face est précédée d'une ligne avec le mot-clé FACE et un numéro, puis chaque point est sur une ligne, sous la forme X:Y (coordonnées 2D pour profil de découpe par fil chaud).\n\n* Export au format DXF avec conservation des entités Arc et Cercle (SketchUp => DXF => usinage):\n--------------------------------------------------------------------------------------------------------------\nLes repères des composants doivent être DIRECTS (règle de la main droite).\n\nL'export concerne uniquement les faces sélectionnées. Si vous ne sélectionnez rien, l'ensemble des faces du modèle seront exportées.\n\nVous pouvez voir en bas à droite de l'écran la progression de l'analyse des faces, tant que le nombre augmente - et que vous n'avez pas eu le message final vous indiquant l'emplacement du fichier - ce n'est pas terminé.\n\nL'export se fait en respectant les coordonnées 3D, attention à la position de vos faces dans l'espace.\n\nL'export avec Z=0 traite uniquement les faces qui sont dans la plan.....Z=0! Cela permet d'exporter le contour de pièce à usiner (laser, fraisage, etc.) sans éclater les composants, et donc de répercuter automatiquement les modifications des composants dans le fichier de sortie pour l'usinage.\nJe sais, c'est dingue!!!\nPour cela il faut mettre l'ensemble de votre projet nettement au-dessus du plan Z=0, puis dupliquer les composants à usiner et aligner toutes les faces des copies sur le plan Z=0 (le plugin ''Throw To'' peut vous aider!). Si vous avez bien utilisé des composants (pas des groupes!), une modification dans le dessin d'ensemble sera automatiquement répercutée sur les copies, et donc sur l'export.\n\nATTENTION : SketchUp représente les arcs et les cercles par des SEGMENTS. L'arc (ou le cercle) RÉEL passe par les EXTRÉMITÉS de ces segments. SI vous COUPEZ un arc ENTRE les extrémités d'un segment, vous créez un point qui est À CÔTÉ de l'arc (ou du cercle) REÉL et si vous connectez ensuite un autre arc ou segment à ce point, LE TRAJET NE SERA PAS CONTINU dans le fichier DXF. Normal, vu que l'entité suivante démarre à côté de l'arc. La qualité de votre fichier DXF dépendra donc de la qualité de votre dessin dans SketchUp! SketchUp Make possède tous les outils pour éviter ce genre de soucis en dessinant correctement.\n\n**************\n\nHistorique et crédits de la partie DXF :\n\nJ'ai initié le codage de ce plugin parce que j'en avait marre de redessiner tous les cercles et arcs en DAO avant de passer à l'usinage quand j'utilisait le plugin skp_to_dxf de Nathan Bromham et Konrad Shroeder (les softs des machines de découpe professionnelles ont leur propre algo d'interpolation circulaire et n'aiment pas les petits segments successifs de SketchUp). J'ai commencé ce projet en étudiant leur plugin - que j'utilisais depuis des années - et donc des parties importantes du code proviennent de leur travail, merci à eux.\nUne fois la décision prise d'améliorer tout ça, et après un coup de pouce de TIG du forum SketchUcation au sujet des ArcCurves, j'ai fait appel à Jérôme David pour un codage à deux (enfin à un et demi, c'est lui qui a quasiment tout fait) de la procédure qui parcoure les boucles externes et internes des faces, remet tout dans l'ordre et identifie les cercles, arcs et polylignes - la première partie bien balèze du truc. Une fois cette base faite, je me suis coltiné tout seul comme un grand la deuxième partie bien balèze, à savoir le transfert des cercles et arcs dans l'OCS du format DXF, avec une routine originale de détermination des angles qui a l'air de marcher à tous les coups (Yeah!). Si les angles ne sont pas respectés, c'est qu'il faut changer le repère local du composant, il n'est pas direct (automatisation dans les ToDo, si vous avez des idées...). J'ai ensuite implémenté l'exploration récursive des groupes et composants et le transfert dans le repère global à partir du code de Nathan et Konrad. J'ai ensuite eu l'idée de l'outil Z=0 que j'ai implémenté, puis j'ai codé les finitions avec (à nouveau!) un coup de pouce de TIG, et les deux langues français/anglais en fonction de la langue d'installation de votre version de SketchUp (ce que vous ne pouvez pas voir, et donc je vous le dis!).\nN'hésitez pas à nous remercier, payer, inviter, acheter une machine... Personnellement je suis très sensible à l'argent et à la flatterie. ;-)\n                                     -- Renaud Iltis, www.cncfab.fr --"
		 
		UI.messagebox help_text, MB_MULTILINE, "CNCFAB plugin v1.2"
	end
	
	def self.readmeenglish
		help_text = "Plugin for import and export by CNCFAB (www.cncfab.fr), manufacturer of CNC hot wire polystyrene cutting machines. \n==========================================================\n\n* Import faces or polylines from TXT (FilChaudNX => SketchUp):\n-----------------------------------------------------------------------------\nThis tool allow to import profiles and projects exported for SketchUp with FilChaudNX hot wire cutting software as a text file (.txt).\n\nIt opens a .txt file in which each line contains the coordinates (x, y, z) of a point, in the form X:Y:Z (attention, decimal separator: the point).\n\nTo define a face, indicate the keyword FACE on a line, then all the points of the outline (one per line), and ends with the keyword END_FACE.\n\nTo reverse the orientation of a face, the end keyword is END_FACE_REVERSE (in SketchUp, the faces created in the plane Z=0 are always facing down ; for the others, it depends of the rotation of the points around the face, clockwise = back face).\n\nFor a polyline, indicate the keyword POLYLINE, then all the points which constitute it (one per line), and ends with the keyword END_POLYLINE.\n\nThe numbers do not have a unit, they will be at the unit of the SketchUp project at the moment of the import (if you import 56.3 in a SketchUp project in mm, it will be 56.3mm, if it is a project in inches, that will be 56.3inches).\n\n* Import polyline from DAT:\n-----------------------------------------------------------------------------\nThis tool is used to import a wing profile file in standard .DAT format, which will be represented by a succession of segments with a total chord of 100 units (multiplication by 100 of the coordinates contained in the .DAT file).\n\n* Export selected faces to X: Y file (SketchUp => FilChaudNX/FilChaud2D):\n-----------------------------------------------------------------------------------------\nThis tool allow to export the 2D coordinates of the points of the outline of the selected faces, then import them into the FilChaudNX and FilChaud2D softwares.\n\nEach face is preceded by a line with the keyword FACE and a number, then each point is on a line, in the form X:Y (2D coordinates for hot wire cutting profiles).\n\n* Export to DXF format with conservation of Arc and Circle entities (SketchUp => DXF => machining):\n--------------------------------------------------------------------------------------------------------------\nThe component's reference axis system are DIRECT (right hand rule).\n\nOnly selected faces are exported. If you don't select anything, it exports the whole model faces.\n\nThere's a progress indicator at the bottom right corner of the window that you can use to monitor the status of the analysis. The indicator will keep increasing until the process is done, with then a message showing the location of the exported file.\n\nThe export is done in 3D space. Pay attention to the space position of your faces.\n\nWith the Z=0 option, the export only processes the faces that are in the...Z=0 plane! This can be used to export parts outlines for machining (e.g. laser, milling) without exploding the components. This way you can modify your components and have the update propagated correctly in the output file.\nCrazy, right?!\nIn order to do this, place the whole model upper than the Z=0 plane, then first duplicate the components, then align all the faces to the Z=0 plane on the duplicates (the ''Throw To'' plugin can be helpfull). If you have used components as recommended (rather than groups), changes in the main design will automatically be propagated to the copies and to the export.\n\nWARNING: SketchUp represents arcs and circles by SEGMENTS. The REAL arc (or circle) goes THROUGH THE VERTICES of these segments. If you CUT an arc INSIDE a segment, you create a point which is NOT ON THE ARC. So if you use this new point to connect another arc or segment, the path will be continuous in SketchUp, but not in the DXF file. It's normal, because the next entity is not starting at the real end of the arc. So the quality of your DXF file depend on the quality of your drawing in SketchUp! SketchUp Make has all the needed tools to solve these problems by drawing a clean model.\n\n-----\n\nHistory and credits of the DXF part :\n\nI started coding this plugin because I was tired of redrawing all circles and arcs in CAD before moving to machining when I used the plugin skp_to_dxf from Nathan Bromham and Konrad Schroeder (the professional cutting machines softwares have their own circular interpolation algorithm and do not like small successive segments of SketchUp). I started by studying their plugin - I had been using it for years - and therefore significant parts of the code is derived from their work. Thank you to them.\nAfter I made the decision to improve all this, I got some help from TIG from the SketchUcation forum regarding the ArcCurves. I then turned to Jerome David for help with coding the procedure that goes through the outer and inner loops, puts everything in order and identifies circles, arcs and polylines. This was the first tricky part for me. Once that was done, I went and coded the second tricky part all on my own : transposing the coordinates of the circles and arcs to the OCS for the DXF format. I came up with a way to calculate the angles that seems to work every time (Yeah!) If the angles are not respected, it is necessary to change the local reference axis of the component, it is not direct (automation is in the ToDo, if you have ideas ...). After that, I implemented the recursive exploration of groups and components, as well as the conversion to the global coordinate system, with the help Konrad and Nathan's code. Once that was done, the idea for the Z=0 tool came to me, and I implemented it. I then tidied up the code with (once again!) some help from TIG and add the language detection (french and english are supported), based on the Sketchup install (you can't see it, so I am telling you about it).\nFeel free to thank us, pay us or invite us, buy a hotwire machine... Myself, I am very sensitive to flattery and monetary donations. ;-)\n                                    -- Renaud Iltis, www.cncfab.fr --"
		UI.messagebox help_text, MB_MULTILINE, "CNCFAB plugin v1.2"
	end

	def self.write_header(file, selection)
		 file.puts(%Q[ 0\nSECTION\n 2\nENTITIES])
	end

	def self.write_footer(file, selection)
		 file.puts(%Q[ 0\nENDSEC\n 0\nEOF])
	end

	# Recursively export faces and exploding groups and components
	def self.write_content(file, selection, tform, z)  # if z==1, it's a 3D export file, if z==0, it's a 2D export file (all the Z values are equal to "0.0" and the normal.z is equal to "1.0"
		 selection.each do |entity|    
				if entity.is_a?(Sketchup::Face)
				transform_vector_of_group(entity.normal, tform)
					 if z == 1 || (z == 0 && transform_vector_of_group(entity.normal, tform).parallel?(Z_AXIS) && transform_point_of_group(entity.vertices[0].position, tform).z.abs < 0.001 ) # Sketchup tolerance : 0.0001 inch
					# if z == 1 || (z == 0 && entity.normal.parallel?(Z_AXIS) && entity.vertices[0].position.z.abs < 0.0004 ) # Sketchup tolerance : 0.0001 inch
							$number_of_exported_faces +=1
							write_face(file, entity, tform, z)
							result = Sketchup.set_status_text $number_of_exported_faces.to_s, SB_VCB_VALUE
					 end
				elsif entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
					 if entity.is_a?(Sketchup::Group)
							definition = entity.entities.parent
					 else
							definition = entity.definition
					 end
					 write_content(file, definition.entities, tform * entity.transformation, z)
				end
		 end
	end

	def self.write_face(file, face, tform, z)
		
			p '***********************'
			
		 face.loops.each do |loop|
				points = []
				last = nil
				last_position = nil
				edges = []
				curves = []
				loop.edges.each do |edge|
					 edges << edge
				end
				edges.each do |edge|
					 last_curve = last.nil? ? last : last.curve
					 curve = edge.curve
					 is_arc = curve.is_a?(Sketchup::ArcCurve)
					 last_was_arc = last_curve.is_a?(Sketchup::ArcCurve)
					 edge_start = edge.start.position
					 edge_end = edge.end.position
					 # test of the orientation of the vertices vs the orientation of the loop
					 if last_position  # not the first edge
							if edge_end == last_position
								 #p 'reverse'
								 edge_end = edge_start
								 edge_start = last_position
							end
					 elsif edges.length > 1  # for the first edge
							if edge_start == edges[1].start.position || edge_start == edges[1].end.position
								 #p 'reverse'
								 edge_start = edge_end
								 edge_end = edge.start.position
							end
					 end
					 if is_arc
							if last_curve == curve
								 #p 'skip'
								 last = edge
								 last_position = edge_end
								 next
							end
							#*********
							# find the center of the curve in OCS, and the normal aZ
							# first, we must move the groups and components entities in the WCS of Sketchup
							aZ = transform_vector_of_group(curve.normal, tform)
							aZ.normalize
							wY = [0.0, 1.0, 0.0]
							wZ = [0.0, 0.0, 1.0]
							if aZ.x.abs < 1.0/64.0 && aZ.y.abs < 1.0/64.0  #warning, if write only "1", instead of "1.0", the result of 1/64 is 0
								 aX = wY.cross aZ
							else
								 aX = wZ.cross aZ
							end
							aX.normalize
							aY = aZ.cross aX
							aY.normalize
							# OCS = (aX,aY,aZ) with [0,0,0] for the origin
							t_ocs2wcs = Geom::Transformation.axes [0.0, 0.0, 0.0], aX, aY, aZ
							t_wcs2ocs = t_ocs2wcs.inverse
							center_in_OCS = transform_point_of_group(curve.center, tform).transform t_wcs2ocs
							#
							#********
							if last_was_arc
								 last = edge
								 last_position = edge_end
								 mod = (curve.end_angle - curve.start_angle).abs.modulo(2 * Math::PI)
								 if mod.to_s == '0.0'
										#p 'circle'
										if curves.include? curve
										#p 'ignore'
										else
											 curves << curve
											 if z == 0 && center_in_OCS.z.abs < 0.1  #with small holes, the value can be between 0.01 and 0.02 after the calculation when the face is in the z plane in SketchUp.
													center_in_OCS.z = 0.0                # but if we are using the option z=0, the value must be set to zero.
											 end
											 if z==1 || (z == 0 && center_in_OCS.z.abs < 0.001 && aZ.parallel?(Z_AXIS))
													write_arc_or_circle('circle', file, center_in_OCS, aZ, curve.radius, z, 0.0, 0.0)
											 end
										end
								 else
										#p 'arc'
										if curves.include? curve
											 #p 'ignore'
										else
											 curves << curve
											 if z == 0 && center_in_OCS.z.abs < 0.1  #with small holes, the value can be between 0.01 and 0.02 after the calculation when the face is in the z plane in SketchUp.
													center_in_OCS.z = 0.0                # but if we are using the option z=0, the value must be set to zero.
											 end
											 if z==1 || (z == 0 && center_in_OCS.z.abs < 0.001 && aZ.parallel?(Z_AXIS))
													arc_angles = []
													arc_angles = define_arc_angles(curve, center_in_OCS, t_wcs2ocs, tform)
													write_arc_or_circle('arc', file, center_in_OCS, aZ, curve.radius, z, arc_angles[0], arc_angles[1])
											 end
										end
								 end
								 next
							end
							if last
								 if points.length > 0
										points << transform_point_of_group(edge_start, tform)
										#p 'poly'
										if z==1 || (z==0 && points.all? {|p| p.z.abs < 0.001})
											write_points(file, points, z)
										end
								 points = []
								 end
							end
							last = edge
							last_position = edge_end
							mod = (curve.end_angle - curve.start_angle).abs.modulo(2 * Math::PI)
							if mod.to_s == '0.0'
								 #p 'circle'
								 if curves.include? curve
										#p 'ignore'
								 else
										if z == 0 && center_in_OCS.z.abs < 0.1  #with small holes, the value can be between 0.01 and 0.02 after the calculation when the face is in the z plane in SketchUp.
													center_in_OCS.z = 0.0                # but if we are using the option z=0, the value must be set to zero.
										end
										curves << curve
										if z==1 || (z == 0 && center_in_OCS.z.abs < 0.001 && aZ.parallel?(Z_AXIS))
											 write_arc_or_circle('circle', file, center_in_OCS, aZ, curve.radius, z, 0.0, 0.0)
										end
								 end
							else
								 #p 'arc'
								 if curves.include? curve
										#p 'ignore'
								 else
										if z == 0 && center_in_OCS.z.abs < 0.1  #with small holes, the value can be between 0.01 and 0.02 after the calculation when the face is in the z plane in SketchUp.
													center_in_OCS.z = 0.0                # but if we are using the option z=0, the value must be set to zero.
										end
										curves << curve
										if z==1 || (z == 0 && center_in_OCS.z.abs < 0.001 && aZ.parallel?(Z_AXIS))
											 arc_angles = []
											 arc_angles = define_arc_angles(curve, center_in_OCS, t_wcs2ocs, tform)
											 write_arc_or_circle('arc', file, center_in_OCS, aZ, curve.radius, z, arc_angles[0], arc_angles[1])
										end
								 end
							end
							next
					 end
					 #p '+'
					 last = edge
					 last_position = edge_end
					 points << transform_point_of_group(edge_start, tform)
				end
				if points.length > 0
					 points << transform_point_of_group(last_position, tform)
					 #p 'poly'
					 if z==1 || (z==0 && points.all? {|p| p.z.abs < 0.001})
							write_points(file, points, z)
					 end
				end
		 end
	end

	def self.define_arc_angles(curve, center_in_OCS, t_wcs2ocs, tform)
		 start_point_arc = []
		 i = nil
		 curve_center = transform_point_of_group(curve.center, tform)
		 curve_xaxis = transform_vector_of_group(curve.xaxis, tform)
		 for i in 0..2
				start_point_arc[i] = curve_center[i].to_f * $faces2dxf_conv + curve_xaxis[i].to_f * $faces2dxf_conv 
		 end
		 curve_normal =  transform_vector_of_group(curve.normal, tform)
		 rotation_arc = Geom::Transformation.rotation curve_center, curve_normal, curve.end_angle
		 end_axis = curve_xaxis.transform rotation_arc
		 end_point_arc = []
		 for i in 0..2
				end_point_arc[i] = curve_center[i].to_f * $faces2dxf_conv + end_axis[i].to_f * $faces2dxf_conv
		 end
		 translation_origin = Geom::Vector3d.new -center_in_OCS[0].to_f* $faces2dxf_conv, -center_in_OCS[1].to_f* $faces2dxf_conv, -center_in_OCS[2].to_f* $faces2dxf_conv
		 start_point_arc_in_OCS = start_point_arc.transform t_wcs2ocs
		 start_point_arc_in_OCS = start_point_arc_in_OCS.transform translation_origin
		 end_point_arc_in_OCS = end_point_arc.transform t_wcs2ocs
		 end_point_arc_in_OCS = end_point_arc_in_OCS.transform translation_origin
		 # it's now a planar situation, and we can use the formulas of the polar coordinates (see the omega function and Wikipedia!)
		 arc_angles = []
		 arc_angles = [omega(start_point_arc_in_OCS), omega(end_point_arc_in_OCS)]
		 return arc_angles 
	end

	def self.transform_vector_of_group(vector, tform)
		 vector_wcs = Geom::Vector3d.new(vector.x, vector.y, vector.z)
		 vector_wcs.transform! tform
		 return vector_wcs
	end

	def self.transform_point_of_group(vertex, tform)
		 point = Geom::Point3d.new(vertex.x, vertex.y, vertex.z)
		 point.transform! tform
		 return point
	end

	def self.omega(point)  #return the oriented angle in polar coordinates
		 x = point[0].to_f
		 y = point[1].to_f
		 if x > 0.0 && y >= 0.0
				o = Math.atan(y / x)
		 elsif x > 0.0 && y < 0.0
				o = Math.atan(y / x) + 2 * Math::PI
		 elsif x < 0.0
				o = Math.atan(y / x) + Math::PI
		 elsif x = 0.0 && y > 0.0
				o = Math::PI / 2
		 elsif x = 0.0 && y < 0.0
				o = 3 * Math::PI / 2
		 end
		 return o.radians.to_f
	end

	def self.write_arc_or_circle(arc_or_circle, file, center, normal, radius, z, start_angle, end_angle)
		 layername="Export SketchUp"
		 arc_or_circle == 'arc' ? file.puts(%Q[  0\nARC\n 8\n#{layername}\n]) : file.puts(%Q[  0\nCIRCLE\n 8\n#{layername}\n])
		 # 10 : x_center
		 if normal.z.to_f < 0 && (z == 0 || (z == 1 && normal.parallel?(Z_AXIS))) # The 2D dxf plugins are not reading the 210,220,230 values and give wrong result if normal.z<0.
				file.puts(%Q[10\n#{(-(center[0].to_f) * $faces2dxf_conv).to_s}]) # In the case normal.z=-1.0, we write z=1.0 and we change x_center, start_angle and end_angle to allow these plugin to give a good result. z_center will also be changed for the 3D viewers, because the OCS is inverted.
		 else
				file.puts(%Q[10\n#{(center[0].to_f * $faces2dxf_conv).to_s}]) 
		 end
		 # 20 : y_center
		 file.puts(%Q[20\n#{(center[1].to_f * $faces2dxf_conv).to_s}])
		 # 30 : z_center
		 if z == 1
				if normal.parallel?(Z_AXIS) && normal.z.to_f < 0
					 file.puts(%Q[30\n#{(-center[2].to_f * $faces2dxf_conv).to_s}])
				else
					 file.puts(%Q[30\n#{(center[2].to_f * $faces2dxf_conv).to_s}])
				end
		 else
				file.puts(%Q[30\n#{"0.0"}])
		 end
		 # 40 : radius
		 file.puts(%Q[40\n#{(radius.to_f * $faces2dxf_conv).to_s}])
		 if arc_or_circle == 'arc'
				# 50, 51 : start_angle & end_angle in OCS
				if normal.z.to_f < 0 && (z == 0 || (z == 1 && normal.parallel?(Z_AXIS)))
					 file.puts(%Q[50\n#{(180.0-end_angle.to_f).to_s}]) # in the case normal.z=-1.0, we change start_angle
					 file.puts(%Q[51\n#{(180.0-start_angle.to_f).to_s}]) # in the case normal.z=-1.0, we change end_angle
				else
					 file.puts(%Q[50\n#{start_angle}])
					 file.puts(%Q[51\n#{end_angle}])
				end
		 end
		 # 210, 220 : x_normal, y_normal
		 z == 1 ? file.puts(%Q[210\n#{normal.x.to_f.to_s}]) : file.puts(%Q[210\n#{"0.0"}])
		 z == 1 ? file.puts(%Q[220\n#{normal.y.to_f.to_s}]) : file.puts(%Q[220\n#{"0.0"}])
		 # 230 : z_normal
		 if z == 0 || (z == 1 && normal.parallel?(Z_AXIS))
				file.puts(%Q[230\n#{"1.0"}]) # in the case normal.z=-1.0, we write z=1.0 and we change xcenter, start_angle and end_angle for the 2D plugins which are not reading the 210,220,230 values
		 else
				file.puts(%Q[230\n#{normal.z.to_f.to_s}])
		 end
	end

	def self.write_points(file, points, z)
		layername="Export SketchUp"
		file.puts(%Q[  0\nPOLYLINE\n 8\n#{layername}\n 66\n     1])
		file.puts(%Q[70\n    8\n 10\n 0.0\n 20\n 0.0\n 30\n 0.0])
		points.each do |pt|
			file.puts(%Q[  0\nVERTEX\n  8\nMY3DLAYER])
			file.puts(%Q[10\n#{(pt.x.to_f * $faces2dxf_conv).to_s}])
			file.puts(%Q[20\n#{(pt.y.to_f * $faces2dxf_conv).to_s}])
			z == 1 ? file.puts(%Q[30\n#{(pt.z.to_f * $faces2dxf_conv).to_s}]) : file.puts(%Q[30\n#{"0.0"}])
		end
		if points.length > 0
			file.puts(%Q[  0\nSEQEND])
		end
	end
	# *****************************************************
	# ****** Import faces and polylines from TXT file *****
	# *****************************************************
	def self.import_faces_and_polylines()
		model = Sketchup.active_model
		cu=Sketchup.active_model.options['UnitsOptions']["LengthUnit"]
		case cu
		when 4  # m
			$TXT2FacesLines_conv=0.0254
		when 3  # cm
			$TXT2FacesLines_conv=2.54
		when 2  # mm
			$TXT2FacesLines_conv=25.4
		when 1  # feet
			$TXT2FacesLines_conv=0.0833333333333333
		when 0  # inches
			$TXT2FacesLines_conv=1.0
		end
		filename = UI.openpanel("Select TXT File", "~", "Text Files|*.txt;||")
		data = File.open(filename,"r")
		while(t = data.gets)
			if t.chomp == "face" or t.chomp == "FACE"
				pts = []
				i = 0
				flag = true
				while(flag == true)
					t = data.gets
					if t.chomp != "end_face" and t.chomp != "END_FACE" and t.chomp != "end_face_reverse" and t.chomp != "END_FACE_REVERSE"
						x,y,z = t.chomp.split(':')
						x = x.to_f / $TXT2FacesLines_conv
						y = y.to_f / $TXT2FacesLines_conv
						z = z.to_f / $TXT2FacesLines_conv
						pts[i] = [x, y, z]
						i = i + 1
					else
						flag = false
					end
				end
				face = model.entities.add_face(pts)
				if t.chomp == "end_face_reverse" or t.chomp == "END_FACE_REVERSE"
					status = face.reverse!
				end
			elsif t.chomp == "polyline" or t.chomp == "POLYLINE"
				pts = []
				i = 0
				flag = true
				while(flag == true)
					t = data.gets
					if t.chomp != "end_polyline" and t.chomp != "END_POLYLINE"
						x,y,z = t.chomp.split(':')
						x = x.to_f / $TXT2FacesLines_conv
						y = y.to_f / $TXT2FacesLines_conv
						z = z.to_f / $TXT2FacesLines_conv
						pts[i] = [x, y, z]
						i = i + 1
					else
						flag = false
					end
				end
				i = i - 2
				if i >=0
					for j in (0..i)
						model.entities.add_edges pts[j], pts[j+1]
					end
				end
			end
		end
	end

	# *****************************************************
	# ****** Import polylines from DAT file *****
	# *****************************************************

def self.import_polylines_from_dat()
  model = Sketchup.active_model
  entities = model.active_entities
	cu=Sketchup.active_model.options['UnitsOptions']["LengthUnit"]
	case cu
	when 4  # m
		$DAT2Lines_conv=0.0254
	when 3  # cm
		$DAT2Lines_conv=2.54
	when 2  # mm
		$DAT2Lines_conv=25.4
	when 1  # feet
		$DAT2Lines_conv=0.0833333333333333
	when 0  # inches
		$DAT2Lines_conv=1.0
	end

  # Ouvrir un dialogue pour sélectionner un fichier .dat
  filename = UI.openpanel("Select DAT File", "~", "DAT Files|*.dat;||")
  return unless filename

  data = File.open(filename, "r")
  points = []

  # Lire la première ligne
  first_line = data.gets.strip
  first_line_coordinates = first_line.split.map(&:to_f)
  
  # Vérifier si la première ligne contient exactement deux coordonnées
  if first_line_coordinates.size == 2
    x, y = first_line_coordinates
    points << [x * 100 / $DAT2Lines_conv, 0, y * 100 / $DAT2Lines_conv]
  end

  # Lire le reste du fichier ligne par ligne
  while (line = data.gets)
    line.strip!
    next if line.empty?  # Ignorer les lignes vides
    
    # Vérifier si la ligne contient exactement deux valeurs après split
    coordinates = line.split.map(&:to_f)
    next unless coordinates.size == 2
    
    x, y = coordinates
    points << [x * 100 / $DAT2Lines_conv, 0, y * 100 / $DAT2Lines_conv]  # Multiplier par 100 et définir z à 0
  end

  data.close

  # Ajouter des lignes reliant les points à l'intérieur du groupe
  (points.size - 1).times do |i|
    model.entities.add_line(points[i], points[i + 1])
  end
end



	# **********************************************************
	# ****** Export faces to TXT file with X:Y coordinates *****
	# **********************************************************
	def self.export_face_points
		selection = Sketchup.active_model.selection
		cu=Sketchup.active_model.options['UnitsOptions']["LengthUnit"]
		case cu
		when 4  # m
			$Faces2XY_conv=0.0254
		when 3  # cm
			$Faces2XY_conv=2.54
		when 2  # mm
			$Faces2XY_conv=25.4
		when 1  # feet
			$Faces2XY_conv=0.0833333333333333
		when 0  # inches
			$Faces2XY_conv=1.0
		end
		if selection.empty?
		UI.messagebox("Nothing selected, export canceled")
		else
			# Count how many faces are in the current selection => must be > 0
		 face_count = 0
		 # Look at all of the entities in the selection.
		 selection.each { |entity|
			 if entity.is_a? Sketchup::Face
			 face_count = face_count + 1
			 end
		 }
		 if face_count > 0
				# Ask for the file path.
				filepath = UI.savepanel("Export selected faces",nil,"*.txt")
				# Don't continue when the user cancelled
				return if filepath.nil?
				# Test of the file extension
				filepath = filepath.tr("\\","/")#in case any PC ruby weirdness...
				if File.basename(filepath,".*") == File.basename(filepath) #the suffix is right or missing
				filepath=File.join(File.dirname(filepath), File.basename(filepath,".txt") + ".txt") #if the suffix is missing or right
				else
				if File.basename(filepath,".*") + ".txt" != File.basename(filepath)#the suffix is different than .txt
					 answer = UI.messagebox("Filename has not the '.txt' extension. Change your extension?", MB_YESNO)
					 if( answer == 6 )
						filepath=File.join(File.dirname(filepath), File.basename(filepath,".*") + ".txt")
					 end
				else
					filepath=File.dirname(filepath) + "/" + File.basename(filepath,".txt") + ".txt" #if the suffix is missing
				end
				end
				
			begin
				# Open a file for writing
				File.open(filepath, "w"){ |file|
				selection = Sketchup.active_model.selection
				# Get an array of faces that are in the selection.
				faces = selection.grep(Sketchup::Face) 
				faces.each_with_index{ |face, index|
					# Write a label for the face.
					file.puts("FACE#{index+1}")
					# Get a transformation object that translates from model space to 2d space of the face.
					t = Geom::Transformation.axes(face.vertices.first.position, *face.normal.axes).inverse
					# Write all vertices to the file.
					blnFirstPoint = true
					first_point_u=0
					first_point_v=0
						face.outer_loop.vertices.each{ |vertex|
					# Get the point of the vertex and apply the transformation.
					point = vertex.position.transform(t)
					# Convert the coordinates
					u, v = point.to_a.map{ |c| c.to_f }
									if blnFirstPoint==true
						first_point_u = u
						first_point_v = v
						blnFirstPoint = false
					end #if
					# Write the coordinates to the file.
					file.puts("#{(u*10000*$Faces2XY_conv).to_i.to_f/10000}:#{(v*10000*$Faces2XY_conv).to_i.to_f/10000}")
								} 
					file.puts("#{(first_point_u*10000*$Faces2XY_conv).to_i.to_f/10000}:#{(first_point_v*10000*$Faces2XY_conv).to_i.to_f/10000}")
					}
				 }
				UI.messagebox("The file is here :" + filepath)
			rescue SystemCallError => e
				if e.message =~ /(No such file or directory)/
					UI.messagebox("Error, the file was not created. Be careful if you are using an older version of SketchUp, the full file path must not contain special characters. The full file path you requested is " + filepath)
				else
					fail() #re-raise the last exception
				end
			end
			else
			UI.messagebox("No faces in the selection, export canceled")
			end
		 end 
	end				
				
	# This will run only once when the file is loaded the first time.
	unless file_loaded?(__FILE__)
		 # Add the tools to the menu.
		 CNCFAB_submenu = UI.menu("Plugins").add_submenu("CNCFAB")

		 cmd_Import_Faces_and_Polylines = UI::Command.new("Import faces or polylines from TXT"){self.import_faces_and_polylines()}
		 cmd_Import_Faces_and_Polylines.small_icon = "CNCFAB_ImportExport/txt2faces_and_polylines_icon_16.png"
		 cmd_Import_Faces_and_Polylines.large_icon = "CNCFAB_ImportExport/txt2faces_and_polylines_icon_24.png"
		 cmd_Import_Faces_and_Polylines.tooltip    = "Import faces or polylines from TXT"
		 cmd_Import_Faces_and_Polylines.status_bar_text = "Explanations in Help menu"
		 cmd_Import_Faces_and_Polylines.menu_text  = "Import faces or polylines from TXT"

		 CNCFAB_submenu.add_item cmd_Import_Faces_and_Polylines
		 
		 cmd_Import_Polylines_from_DAT = UI::Command.new("Import polylines from DAT") { self.import_polylines_from_dat() }
		 cmd_Import_Polylines_from_DAT.small_icon = "CNCFAB_ImportExport/dat2polylines_icon_16.png"
		 cmd_Import_Polylines_from_DAT.large_icon = "CNCFAB_ImportExport/dat2polylines_icon_24.png"
		 cmd_Import_Polylines_from_DAT.tooltip = "Import polylines from DAT"
		 cmd_Import_Polylines_from_DAT.status_bar_text = "Explanations in Help menu"
		 cmd_Import_Polylines_from_DAT.menu_text = "Import polylines from DAT"

		 CNCFAB_submenu.add_item cmd_Import_Polylines_from_DAT
		 
		 cmd_Export_Faces_to_XY_file = UI::Command.new("Export selected faces to X:Y file"){self.export_face_points()}
		 cmd_Export_Faces_to_XY_file.small_icon = "CNCFAB_ImportExport/faces2xy_icon_16.png"
		 cmd_Export_Faces_to_XY_file.large_icon = "CNCFAB_ImportExport/faces2xy_icon_24.png"
		 cmd_Export_Faces_to_XY_file.tooltip    = "Export selected faces to X:Y file"
		 cmd_Export_Faces_to_XY_file.status_bar_text = "Explanations in Help menu"
		 cmd_Export_Faces_to_XY_file.menu_text  = "Export selected faces to X:Y file"

		 CNCFAB_submenu.add_item cmd_Export_Faces_to_XY_file
		 
		 cmd_faces2DXF_all_or_selected = UI::Command.new("Export faces to DXF entities (selected or all)"){self.export_faces(1)}
		 cmd_faces2DXF_all_or_selected.small_icon = "CNCFAB_ImportExport/faces2DXFentities_icon_16.png"
		 cmd_faces2DXF_all_or_selected.large_icon = "CNCFAB_ImportExport/faces2DXFentities_icon_24.png"
		 cmd_faces2DXF_all_or_selected.tooltip    = "Export to DXF (Faces selected or all)"
		 cmd_faces2DXF_all_or_selected.status_bar_text = "Explanations in Help menu"
		 cmd_faces2DXF_all_or_selected.menu_text  = "Export faces to DXF entities (selected or all)"

		 CNCFAB_submenu.add_item cmd_faces2DXF_all_or_selected

		 cmd_faces2DXF_only_z0 = UI::Command.new("Export faces to DXF entities (only z=0)"){self.export_faces(0)}
		 cmd_faces2DXF_only_z0.small_icon = "CNCFAB_ImportExport/faces2DXFentities_z0_icon_16.png"
		 cmd_faces2DXF_only_z0.large_icon = "CNCFAB_ImportExport/faces2DXFentities_z0_icon_24.png"
		 cmd_faces2DXF_only_z0.tooltip    = "Export to DXF (faces at z=0)"
		 cmd_faces2DXF_only_z0.status_bar_text = "Explanations in Help menu"
		 cmd_faces2DXF_only_z0.menu_text  = "Export faces to DXF entities (only z=0)"

		 CNCFAB_submenu.add_item cmd_faces2DXF_only_z0

		 CNCFAB_submenu.add_separator
		 menu_CNCFAB_help_french = UI::Command.new("Aide en Français"){self.readmefrench}
		 menu_CNCFAB_help_french.menu_text  = "Aide en Français"
		 CNCFAB_submenu.add_item menu_CNCFAB_help_french
		 
		 menu_CNCFAB_help_english = UI::Command.new("Help in English"){self.readmeenglish}
		 menu_CNCFAB_help_english.menu_text  = "Help in English"
		 CNCFAB_submenu.add_item menu_CNCFAB_help_english
		 
		 toolbar_CNCFAB = UI.toolbar("CNCFAB")
		 toolbar_CNCFAB.add_item(cmd_Import_Faces_and_Polylines)
		 toolbar_CNCFAB.add_item(cmd_Import_Polylines_from_DAT)
		 toolbar_CNCFAB.add_item(cmd_Export_Faces_to_XY_file)
		 toolbar_CNCFAB.add_item(cmd_faces2DXF_all_or_selected)
		 toolbar_CNCFAB.add_item(cmd_faces2DXF_only_z0)
		 if toolbar_CNCFAB.get_last_state == TB_VISIBLE
				UI.start_timer(0.1, false) { toolbar_CNCFAB.restore }
		 elsif toolbar_CNCFAB.get_last_state == TB_NEVER_SHOWN
				toolbar_CNCFAB.show
		 end
	end

	file_loaded(__FILE__)

	end  #module ImportExport
	
end  #module CNCFAB